/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  bootp.c
 * Purpose: Get information for client with BOOTP protocol
 * Entries: bootp
 *
 **************************************************************************
 *
 * Copyright (C) 1995-1999 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************
 *
 * We can print error messages within this module as it gets never called
 * by a PXE function.
 */


#include <general.h>
#include <kernel/net.h>
#include <kernel/arpa.h>
#include <kernel/romlib.h>
#include <pxe/common.h>
#include "arpapriv.h"
#include "bootp.h"



/*
 **************************************************************************
 * 
 * To make it easier to specify the original BOOTP record size use the
 * following define:
 */
#define BOOTP_REC_SIZE	((int)((char *)boot_rec.bp_ext - (char *)(&boot_rec)))


/*
 **************************************************************************
 * 
 * Global variables:
 */
static unsigned long boot_xid;			/* bootp transaction ID	*/
static unsigned long boot_start_time;		/* when started to boot	*/
struct bootp boot_rec;				/* bootp record		*/



/*
 **************************************************************************
 * 
 * Send a BOOTP request via UDP.
 *
 */
static int bootp_send()
{
  /* Assemble the BOOTP request */
  memset(&boot_rec, 0, BOOTP_REC_SIZE);
  memcpy(boot_rec.bp_chaddr, myhwaddr, ETH_ALEN);
  *((unsigned long *)&(boot_rec.bp_vend[0])) = VM_RFC1048;
  boot_rec.bp_op = BOOTP_REQUEST;
  boot_rec.bp_hwtype = BOOTP_ETHER;
  boot_rec.bp_hlen = ETH_ALEN;
  boot_rec.bp_xid = boot_xid;
  boot_rec.bp_secs = htons(((unsigned short)(get_ticks() - boot_start_time)) / 18);
  boot_rec.bp_ciaddr = n_IP_ANY;
  boot_rec.bp_siaddr = n_IP_BROADCAST;

  /* Send the bootp record as a broadcast message to all servers */
  return(udp_write((unsigned char *)(&boot_rec), getds(), BOOTP_REC_SIZE) ==
							PXENV_STATUS_SUCCESS);
}



/*
 **************************************************************************
 * 
 * Get a BOOTP record from the server.
 *
 */
static int bootp_get()
{
  char errch;
  unsigned int timeout;
  register int i;

  /*
   * Setup the IP interface information. Our own address has to be
   * IP_ANY, so that the IP layer will not discard the incoming packets.
   * Then open a UDP socket.
   */
  myipaddr = IP_ANY;
  mynetmask = IP_CLASS_A;
  printf("\n\nBOOTP: ");
  if ((i = udp_open(IP_BROADCAST, IP_ANY,
			BOOTP_C_PORT, BOOTP_S_PORT)) != PXENV_STATUS_SUCCESS) {
	printf("udp_open error %x\n", i);
	return(FALSE);
  }

  /*
   * Now loop until receiving a reply from a server. The retry time is
   * computed as suggested in RFC951.
   */
  timeout = MIN_TIMEOUT + random(0x001f);
  printf("Sending request (press ESC to abort): ");
  while (TRUE) {
	boot_xid = get_ticks() + random(0x7fff);
	if (!bootp_send()) {
		printf("\nBOOTP: Unable to send request\n");
		i = -1;
		break;
	}
	i = udp_read((unsigned char *)(&boot_rec), getds(), BOOTP_REC_SIZE,
							timeout, CHR_ESC);
	if (i < 0 && i != ERR_UDP_TIMEOUT) {	/* user pressed ESC */
		printf("\nBOOTP: Aborted\n");
		break;
	} else if (i == ERR_UDP_TIMEOUT)	/* timeout */
		errch = '.';
	else if (boot_rec.bp_xid != boot_xid)	/* rule out duplicate packet */
		errch = '?';
	else
		break;
	timeout = (timeout << 1) + random(0x001f);
	while (timeout > 60)		/* dont allow more than 60 seconds */
		timeout -= 60;
	printf("%c", errch);
  }

  if (i > 0) {
	if (i < BOOTP_REC_SIZE ||
	    (boot_rec.bp_vend[0] != '\0' &&
	     (*((unsigned long *)&(boot_rec.bp_vend[0])) != VM_RFC1048 ||
	      get_vend(VEND_END) == NULL))) {
		printf("\nBOOTP: Invalid record\n");
		i = -1;
	} else
		printf("ok\n");
  }

  if (udp_close() != PXENV_STATUS_SUCCESS) {
	printf("BOOTP: udp_close error\n");
	i = -1;
  }
  return(i > 0);
}



#ifdef BPDEBUG
/*
 **************************************************************************
 * 
 * Print the complete BOOTP record
 */
static void print_rec()
{
  register unsigned char *cp, *end;
  int i, c, adr;

  adr = 0;
  cp = (unsigned char *)boot_rec;
  if ((end = get_vend(VEND_END)) == NULL)
	end = cp + sizeof(struct bootp);
  printf("\n\n");
  while(cp < end) {
	if (adr < 16) printf("00%x  ", adr);
	else if (adr < 256) printf("0%x  ", adr);
	else printf("%x  ", adr);

	for (i = 0; i < 16 && cp <= end; i++, cp++) {
		c = *cp & 0xff;
		if (c < 16) printf("0%x ", c);
		else printf("%x ", c);
	}
	printf("\n");
	adr += 16;
  }
  (void)getkey(0);
  printf("\n\n");
}
#endif



/*
 **************************************************************************
 * 
 * Handle BOOTP protocol. If we have a BOOTP record already, we just set
 * the network parameters anew.
 */
int bootp()
{
  register unsigned char *cp;

  /* Set boot start time - required for gateway booting */
  boot_start_time = get_ticks();

  /* Get BOOTP record from server */
  if (!bootp_get())
	return(FALSE);

  /* Setup the network interface with the values received from BOOTP server */
  myipaddr = ntohl(boot_rec.bp_yiaddr);
  if ((cp = get_vend(VEND_SUBNET)) != NULL)
	mynetmask = ntohl(*((t_ipaddr *)(cp + 1)));
  else if (IN_CLASS_A(myipaddr))
	mynetmask = IP_CLASS_A;
  else if (IN_CLASS_B(myipaddr))
	mynetmask = IP_CLASS_B;
  else if (IN_CLASS_C(myipaddr))
	mynetmask = IP_CLASS_C;

#ifndef NOBPEXT
  /* Next try to get the extensions file if there is any specified */
  if ((cp = get_vend(VEND_EXTFILE)) != NULL && !get_ext(cp))
	return(FALSE);
#endif

#ifdef BPDEBUG
  print_rec();
#endif

  return(TRUE);
}

