/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  boot.c
 * Purpose: Perform the actual booting process
 * Entries: do_boot
 *
 **************************************************************************
 *
 * Copyright (C) 1995-1999 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <general.h>
#include <kernel/net.h>
#include <kernel/arpa.h>
#include <kernel/romlib.h>
#include <pxe/common.h>
#include <pxe/pxegeneral.h>
#include "bootpriv.h"
#include "menu.h"
#include "load.h"


/*
 **************************************************************************
 * 
 * Variables local to this module
 */
static jmp_buf restartenv;		/* setjmp buffer for restart	*/



/*
 **************************************************************************
 * 
 * Print an IP number. Note that the IP address should be in network
 * order.
 */
static void printip(ip)
t_ipaddr ip;
{
  register unsigned char *cp = (unsigned char *)&ip;

  printf("%u.%u.%u.%u", cp[3], cp[2], cp[1], cp[0]);
}



/*
 **************************************************************************
 * 
 * Set parameters from BOOTP record
 */
static int setparams()
{
  register unsigned char *cp;
  register int len;
  unsigned char *sname;
  int snamlen;
  int overload = 0;

  memset(&ldparams, 0, sizeof(ldparams));

  /* Get the overloading tag */
  if ((cp = get_vend(VEND_OVERLOAD)) != NULL)
	overload = *(cp + 1);

  /* Determine the IP address of the TFTP server */
  if (boot_rec.bp_siaddr == n_IP_ANY) {
	if ((cp = get_vend(VEND_SERVER)) != NULL)
		ldparams.tftp.server = *((t_ipaddr *)(cp + 1));
  } else
	ldparams.tftp.server = boot_rec.bp_siaddr;
  if (ldparams.tftp.server == n_IP_ANY) {
	printf("Missing server IP address\n");
	return(FALSE);
  }

  /* Determine name of server the boot image file will come from */
  if ((cp = get_vend(VEND_TFTPNAME)) != NULL) {
	sname = cp + 1;
	snamlen = *cp;
  } else if (!(overload & VEND_OVR_SNAME)) {
	sname = boot_rec.bp_sname;
	snamlen = sizeof(boot_rec.bp_sname);
  } else
	sname = NULL;

  /* Determine name of boot image file to load */
  if ((cp = get_vend(VEND_BOOTFILE)) != NULL) {
	len = *cp;
	if (len > sizeof(ldparams.tftp.filename))
		len = sizeof(ldparams.tftp.filename);
	memcpy(ldparams.tftp.filename, cp + 1, len);
  } else if (!(overload & VEND_OVR_FILE))
	memcpy(ldparams.tftp.filename, boot_rec.bp_file,
						sizeof(ldparams.tftp.filename));
  if (ldparams.tftp.filename[0] == '\0') {
	printf("Missing boot file name\n");
	return(FALSE);
  }

  /* Get any router information */
  if ((cp = get_vend(VEND_ROUTER)) != NULL)
	ldparams.tftp.gateway = *((t_ipaddr *)(cp + 1));

  /* Print some of the information we got so far */
  printf("\n\nLocal IP:  "); printip(myipaddr);
  printf("\nServer IP: "); printip(ntohl(ldparams.tftp.server));
  if (sname != NULL)
	printf("  (%ls)", sname, snamlen);
  printf("\n");
  return(TRUE);
}



/*
 **************************************************************************
 * 
 * Actually perform the booting process by first calling the BOOTP client,
 * and then loading the operating system using TFTP. This routine gets
 * called from the assembler startup code.
 *
 */
void do_boot()
{
  if (bootp()) {
#ifndef NOMENU
	if (domenu() != MENU_ABORT) {
#endif
		while (setparams()) {
			if (load(FALSE) != PXENV_STATUS_SUCCESS)
				break;
			printf("\n\nStarting image...\n");
			setjmp(restartenv);
			if (!exec_image(ldparams.mode,
					ldparams.exec, ldparams.header,
					&boot_rec, ldparams.drive))
				break;
		}
#ifndef NOMENU
	}
#endif
  }
}



/*
 **************************************************************************
 * 
 * Restart TFTP PXE function
 */
int pxe_restart(params)
t_tftp_read_file *params;
{
  register int ret = PXENV_STATUS_FAILURE;

  memset(&ldparams, 0, sizeof(ldparams));
  memcpy(&ldparams.tftp, params, sizeof(ldparams.tftp));
  if ((ret = load(TRUE)) == PXENV_STATUS_SUCCESS)
	/*
	 * We have to use longjump here to jump back to the original
	 * bootrom loader code. Otherwise, the bootrom stack will
	 * overflow with multiple calls of this routine.
	 */
	longjmp(restartenv, 0);

  return(ret);
}



/*
 **************************************************************************
 * 
 * Return BOOTP/DHCP information
 */
int pxe_get_binl_info(params)
t_gen_get_binl_info *params;
{
  register unsigned char *cp;
  unsigned int len, limit, ret;

  /* Determine which packet the caller wants */
  ret = PXENV_STATUS_FAILURE;
  if (params->packet_type == GEN_PACKET_BINL_REPLY) {
	cp = (unsigned char *)&boot_rec;
	limit = sizeof(boot_rec);
  } else {
	cp = NULL;
	limit = 0;
  }

  /* Now return the information */
  if (cp != NULL) {
	len = get_vend(VEND_END) - cp;
	if ((params->buffer_offset + params->buffer_segment +
	     params->buffer_size) == 0) {
		params->buffer_size = len;
		params->buffer_offset = (unsigned int )cp;
		params->buffer_segment = getds();
		params->buffer_limit = limit;
		ret = PXENV_STATUS_SUCCESS;
	} else if (params->buffer_size >= len) {
		fmemcpy(params->buffer_offset, params->buffer_segment, cp, len);
		params->buffer_size = len;
		params->buffer_limit = limit;
		ret = PXENV_STATUS_SUCCESS;
	}
  }
  return(ret);
}

