/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  arp.c
 * Purpose: ARP handling
 * Entries: ip_resolve, set_gateway, arp_poll, arp_recv
 *
 **************************************************************************
 *
 * Copyright (C) 1995-1999 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <general.h>
#include <kernel/net.h>
#include <kernel/romlib.h>
#include <pxe/common.h>
#include <pxe/undi.h>
#include "netpriv.h"
#include "arp.h"



/*
 **************************************************************************
 * 
 * Global variables
 */
static struct arphdr arpbuf;			/* ARP send buffer	*/
static struct arp_cache arp_list[ARPNUM];	/* ARP address cache	*/
static int next_arp;				/* next table entry	*/
static int sendflag;				/* indicates send reqd	*/
static t_ipaddr localip;			/* my local IP address	*/

static unsigned char sendaddr[ETH_ALEN];	/* send address		*/
static unsigned char bcasthw[ETH_ALEN] = {	/* broadcast address	*/
		0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};

static int sizes[] = { sizeof(arpbuf) };
static unsigned char *bufs[] = { (unsigned char *)&arpbuf , NULL };



/*
 **************************************************************************
 * 
 * Find an IP address in the ARP cache. Return -1 if not found.
 */
static int findip(ip)
t_ipaddr ip;
{
  register int i;

  for (i = 0; i < ARPNUM; i++)
	if (arp_list[i].ipaddr == ip)
		return(i);
  return(-1);
}



/*
 **************************************************************************
 * 
 * Add a new entry to the ARP cache.
 */
static void addcache(ha, ip)
unsigned char *ha;
t_ipaddr       ip;
{
  register int i;

  /*
   * If the IP address already exists in the ARP table then just replace
   * it's hardware address with the new onbe. Otherwise create a new ARP
   * table entry.
   */
  if ((i = findip(ip)) < 0) {
	i = next_arp++;
	if (next_arp >= ARPNUM)
		next_arp = 0;
	arp_list[i].ipaddr = ip;
  }
  memcpy(arp_list[i].hwaddr, ha, ETH_ALEN);
}



/*
 **************************************************************************
 * 
 * ARP receiver routine
 */
void arp_recv(framesize)
int framesize;
{
  int bufsize;

  /*
   * Copy received buffer into local ARP buffer, but only if buffer free.
   * Note that an ARP packet will always fit into one buffer (since UNDI
   * has a minimum buffer size of 512), so that we will always get a full
   * ARP buffer.
   */
  if (sendflag)
	goto recv_ret;
  bufsize = undi_get_buf(sizeof(arpbuf), (unsigned char *)&arpbuf);

  /*
   * Do some checks on the received packet. It must be an ARP packet,
   * with an ethernet hardware address, and for the IP protocol. Other
   * packets just get discarded. Then save the sender's address.
   */
  if (bufsize < sizeof(arpbuf) ||
      arpbuf.ar_hrd != htons(ARPHRD_ETHER) || arpbuf.ar_hln != ETH_ALEN ||
      arpbuf.ar_pro != htons(ARPPROT_IP) || arpbuf.ar_pln != IP_ALEN)
	goto recv_ret;

  addcache(arpbuf.ar_sha, _ntohl(arpbuf.ar_sip));

  /* If the packet is a request, send our own hardware address back */
  sendflag = (arpbuf.ar_op == htons(ARPOP_REQUEST) &&
					arpbuf.ar_tip == htonl(localip));
  if (sendflag) {
	arpbuf.ar_op  = htons(ARPOP_REPLY);
	arpbuf.ar_tip = arpbuf.ar_sip;
	arpbuf.ar_sip = htonl(localip);
	memcpy(arpbuf.ar_tha, arpbuf.ar_sha, ETH_ALEN);
	memcpy(arpbuf.ar_sha, myhwaddr, ETH_ALEN);
	memcpy(sendaddr, arpbuf.ar_tha, ETH_ALEN);
  }

recv_ret:
  return;
}



/*
 **************************************************************************
 * 
 * ARP polling routine. This sends a previously setup ARP packet.
 */
int arp_poll()
{
  register int status = PXENV_STATUS_SUCCESS;

  if (sendflag) {
	sendflag = FALSE;
	status = undi_send(sendaddr, UNDI_PROT_ARP, 1, sizes, bufs);
  }
  return(status);
}



/*
 **************************************************************************
 * 
 * Resolve IP address and return pointer to hardware address.
 */
int ip_resolve(myip, ip, gw, hwaddr)
t_ipaddr myip;
t_ipaddr ip;
t_ipaddr gw;
unsigned char **hwaddr;
{
  int i, ret;

#ifdef DEBUG
  printf("ARP: Resolving IP address %lx\n", ip);
#endif

  /*
   * We need to set our own IP address first, so that it's available when we
   * get an ARP request from an external system.
   */
  localip = (myip == IP_ANY ? myipaddr : myip);

  /* Check that the gateway address is correct */
  if (gw != IP_ANY &&
      (gw == IP_BROADCAST || gw == localip || ((gw ^ localip) & mynetmask)))
	gw = IP_ANY;

  /* Clear send queue in case we need it */
  if ((ret = arp_poll()) != PXENV_STATUS_SUCCESS)
	goto res_ret;

  /* Check for invalid IP addresses */
  ret = PXENV_STATUS_FAILURE;
  if (ip == IP_ANY || ip == localip || ip == IP_LOCALHOST)
	goto res_ret;

  /*
   * If the destination is not on subnet, send the packets through gateway.
   * Also check for a local broadcast address.
   */
  if (ip != IP_BROADCAST) {
	/*
	 * This double-if construction lets the compiler produce better
	 * optimized code compared to a &&.
	 */
	if ((ip ^ localip) & mynetmask) {
		if (gw != IP_ANY)
			ip = gw;
		else
			goto res_ret;
	}
  }

  /* Check for broadcast address */
  ret = PXENV_STATUS_SUCCESS;
  if (ip == IP_BROADCAST || !((ip ^ IP_BROADCAST) & ~mynetmask)) {
	memcpy(hwaddr, bcasthw, ETH_ALEN);
	goto res_ret;
  }

  /* Check if we got the address in the cache already */
  if ((i = findip(ip)) >= 0) {
	memcpy(hwaddr, arp_list[i].hwaddr, ETH_ALEN);
	goto res_ret;
  }

  /* No, we have to broadcast a request */
  arpbuf.ar_hrd = htons(ARPHRD_ETHER);
  arpbuf.ar_pro = htons(ARPPROT_IP);
  arpbuf.ar_hln = ETH_ALEN;
  arpbuf.ar_pln = IP_ALEN;
  arpbuf.ar_op  = htons(ARPOP_REQUEST);
  arpbuf.ar_sip = htonl(localip);
  arpbuf.ar_tip = htonl(ip);
  memcpy(arpbuf.ar_sha, myhwaddr, ETH_ALEN);
  memset(arpbuf.ar_tha, 0, ETH_ALEN);
  bufs[1] = getds();
  if ((ret = undi_send(NULL, UNDI_PROT_ARP, 1, sizes, bufs))
						 != PXENV_STATUS_SUCCESS)
	goto res_ret;

#ifdef DEBUG
  printf("ARP: Request sent\n");
#endif

  /* Wait for someone to answer the broadcast request */
  set_timeout(ARP_TIMEOUT);
  while (!chk_timeout() && (i = findip(ip)) < 0) ;
  if (i >= 0)
	memcpy(hwaddr, arp_list[i].hwaddr, ETH_ALEN);
  else
	ret = PXENV_STATUS_ARP_TIMEOUT;

res_ret:
  return(ret);
}

