/*
 **************************************************************************
 *
 * Boot-ROM-Code to load an operating system across a TCP/IP network.
 *
 * Module:  udp.c
 * Purpose: UDP and IP handling
 * Entries: udp_open, udp_close, udp_read, udp_write, udp_recv
 *
 **************************************************************************
 *
 * Copyright (C) 1995-1999 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <general.h>
#include <kernel/net.h>
#include <kernel/romlib.h>
#include <pxe/common.h>
#include <pxe/undi.h>
#include <pxe/udp.h>
#include "netpriv.h"
#include "arp.h"
#include "ip.h"
#include "udp.h"


/*
 **************************************************************************
 * 
 * Public variables:
 */
unsigned char     myhwaddr[ETH_ALEN];		/* my own hardware addr	*/
         t_ipaddr myipaddr;			/* my own IP address	*/
         t_ipaddr mynetmask;			/* my own netmask	*/



/*
 **************************************************************************
 * 
 * Global variables:
 */
static unsigned short usource;		/* Source port number		*/
static unsigned short udest;		/* Destination port number	*/
static t_ipaddr       udaddr;		/* Destination IP address	*/
static t_ipaddr       usaddr;		/* Source IP address		*/
static t_ipaddr       gateway;		/* Gateway IP address		*/
static int            packetid;		/* packet ID counter		*/
static int            buflen;		/* length of contents in buffer	*/
static struct udphdr  udpbuf;		/* UDP header buffer		*/
static struct iphdr   ipbuf;		/* IP header buffer		*/
static unsigned char  readbuf[IP_RECV_SIZE];	/* receive buffer	*/
static unsigned char *readptr;			/* receive buffer ptr	*/



/*
 **************************************************************************
 * 
 * Open a new UDP socket.
 */
int udp_open(daddr, gw, source, dest)
t_ipaddr daddr;
t_ipaddr gw;
int      source;
int      dest;
{
  /* Set global variables */
  usource = source;
  udest = dest;
  usaddr = myipaddr;
  udaddr = daddr;
  gateway = gw;
  buflen = -1;

  /* Open UNDI interface */
  return(undi_open());
}



/*
 **************************************************************************
 * 
 * Close a UDP socket
 */
int udp_close()
{
  return(undi_close());
}



/*
 **************************************************************************
 * 
 * IP receiver routine
 */
void udp_recv(framesize)
int framesize;
{
#define ipp ((struct iphdr *)readbuf)
#define udpp ((struct udphdr *)(readbuf + IP_MIN_HSIZE))

  int size;

  /* If we have a frame size of zero we have to recycle the receive buffer */
  if (framesize == 0) {
	readptr = readbuf;
	goto recv_ret;
  }

  /*
   * Copy received buffer into local IP buffer, but only if buffer free.
   * If we don't have a full frame yet, return immediately after copying
   * the data.
   */
  if (buflen >= 0)
	goto recv_ret;
  size = readbuf + sizeof(readbuf) - readptr;
  if (size > 0)
	size = undi_get_buf(size, readptr);
  readptr += size;
  if ((readptr - readbuf) < framesize)
	goto recv_ret;

  /*
   * Do some checks on the received packet. It must be an IP packet
   * with a valid version number and header size. Also the checksum
   * of the header must be correct, and the packet should not be frag-
   * mented. The packet should also not be too large, and it has to be
   * a UDP packet.
   */
  size = ntohs(ipp->ip_tot_len) - IP_MIN_HSIZE;
  if (framesize <= (IP_MIN_HSIZE + sizeof(struct udphdr)) ||
    size < sizeof(struct udphdr) || size > (IP_RECV_SIZE - IP_MIN_HSIZE) ||
    ipp->ip_ihl_version != ((IP_VERSION << 4) | (IP_MIN_HSIZE >> 2)) ||
    ipchksum((unsigned char *)ipp, IP_MIN_HSIZE) != 0 ||
    (ntohs(ipp->ip_frag_off) & 0x3fff) || ipp->ip_protocol != IPPROTO_UDP ||
    ntohs(udpp->udp_len) > size)
	goto recv_ret;

#ifdef DEBUG
  printf("UDP: got packet from %lx.%d for local port %d\n",
	ntohl(ipp->ip_daddr), ntohs(udpp->udp_source), ntohs(udpp->udp_dest));
#endif

  buflen = ntohs(udpp->udp_len) - sizeof(struct udphdr);

recv_ret:
  return;

#undef ipp
#undef udpp
}



/*
 **************************************************************************
 * 
 * Read one packet from a UDP socket
 */
int udp_read(buf, bufseg, bufsize, timeout, abortch)
unsigned char *buf;
unsigned int   bufseg;
int            bufsize;
int            timeout;
char           abortch;
{
#define ipp ((struct iphdr *)readbuf)
#define udpp ((struct udphdr *)(readbuf + IP_MIN_HSIZE))

  int len;

  /* Wait until we receive a correct packet */
  set_timeout(timeout);
  while (!chk_timeout()) {
	if (abortch && chkkey() == abortch)
		return(ERR_UDP_ABORT);
	if (buflen >= 0) {
		if ((usaddr == IP_ANY || usaddr == ntohl(ipp->ip_daddr)) &&
		    (usource == 0 || usource == ntohs(udpp->udp_dest)))
			break;
		buflen = -1;	/* ignore packet */
	}
  }

  /* Check for timeout and proper buffer size */
  if (buflen < 0)
	return(ERR_UDP_TIMEOUT);
  if ((len = buflen) > bufsize)
	len = bufsize;

  /*
   * Setup some global variables. This is required for TFTP and DHCP.
   */
  udest = ntohs(udpp->udp_source);
  udaddr = ntohl(ipp->ip_saddr);

  /* Copy the buffer and mark it as empty */
  fmemcpy((unsigned int)buf, bufseg,
			(unsigned char *)udpp + sizeof(struct udphdr), len);
  buflen = -1;
  return(len);

#undef ipp
#undef udpp
}



/*
 **************************************************************************
 * 
 * Write one packet to a UDP socket
 */
int udp_write(buf, bufseg, len)
unsigned char *buf;
unsigned int   bufseg;
int            len;
{
  int status;
  unsigned char udhw[ETH_ALEN];
  static int sizes[] = { sizeof(ipbuf), sizeof(udpbuf), 0 };
  static unsigned char *bufs[] = {
		(unsigned char *)&ipbuf,	NULL,
		(unsigned char *)&udpbuf,	NULL,
		NULL, 				NULL};

  /* Determine the hardware address of the recipient */
  if ((status = ip_resolve(usaddr, udaddr, gateway, &udhw)) !=
						PXENV_STATUS_SUCCESS)
	return(status);

  /*
   * Copy the packet into the output buffer. It's size should not be larger
   * than maximum UDP packet
   */
  if (len > UDP_SEND_SIZE)
	len = UDP_SEND_SIZE;
  sizes[2] = len;
  bufs[4] = buf;
  bufs[5] = (unsigned char *)bufseg;
  bufs[1] = bufs[3] = (unsigned char *)getds();

#ifdef DEBUG
  printf("UDP: sending %d bytes from %ld.%d to %ld.%d (%x:%x:%x:%x:%x:%x)\n",
			len, usaddr, usource, udaddr, udest,
			udhw[0], udhw[1], udhw[2], udhw[3], udhw[4], udhw[5]);
#endif

  /* Prepare UDP datagram */
  udpbuf.udp_source = htons(usource);
  udpbuf.udp_dest = htons(udest);
  udpbuf.udp_len = htons(len + sizeof(udpbuf));
  udpbuf.udp_check = 0;

  /* Prepare IP header */
  ipbuf.ip_ihl_version = ((IP_VERSION << 4) | ((sizeof(ipbuf)/4) & 0x0f)) & 0xff;
  ipbuf.ip_tos         = IP_STD_TOS;
  ipbuf.ip_tot_len     = htons(sizeof(ipbuf) + sizeof(udpbuf) + len);
  ipbuf.ip_id          = htons(++packetid);
  ipbuf.ip_frag_off    = 0;
  ipbuf.ip_check       = 0;
  ipbuf.ip_ttl         = IP_TTL;
  ipbuf.ip_protocol    = IPPROTO_UDP;
  ipbuf.ip_daddr       = htonl(udaddr);
  ipbuf.ip_saddr       = htonl(usaddr);
  ipbuf.ip_check       = ipchksum((unsigned char *)&ipbuf, sizeof(ipbuf));

  /* Finally send out the packet */
  return(undi_send(udhw, UNDI_PROT_IP, 3, sizes, bufs));
}



/*
 **************************************************************************
 * 
 * UDP open PXE API function
 */
static int pxe_udp_open(params)
t_udp_open *params;
{
  if (params->src_ip == IP_ANY)
	params->src_ip = myipaddr;
  usaddr = params->src_ip;
  buflen = -1;
  return(undi_open());
}



/*
 **************************************************************************
 * 
 * UDP close PXE API function
 */
static int pxe_udp_close(params)
t_udp_close *params;
{
  return(undi_close());
}



/*
 **************************************************************************
 * 
 * UDP read PXE API function
 */
static int pxe_udp_read(params)
t_udp_read *params;
{
  int status;
  t_ipaddr oldsaddr = usaddr;

  /* Set global variables for this call only */
  usource = params->d_port;
  udest = -1;
  usaddr = params->dest_ip;
  udaddr = IP_ANY;

  /* Wait for a packet to come in */
  status = udp_read(params->bufofs, params->bufseg, params->bufsize,
							UDP_TIMEOUT, 0);
  if (status == ERR_UDP_TIMEOUT)
	status = PXENV_STATUS_ARP_TIMEOUT;
  else if (status == ERR_UDP_ABORT)
	status = PXENV_STATUS_ARP_CANCELED;
  else if (status < 0)
	status = PXENV_STATUS_FAILURE;
  else {
	params->src_ip = udaddr;
	params->s_port = udest;
	status = PXENV_STATUS_SUCCESS;
  }
  usaddr = oldsaddr;
  return(status);
}



/*
 **************************************************************************
 * 
 * UDP write PXE API function
 */
static int pxe_udp_write(params)
t_udp_write *params;
{
  usource = params->src_port;
  udest = params->dest_port;
  udaddr = params->ip;
  gateway = params->gw;
  return(udp_write(params->bufofs, params->bufseg, params->bufsize));
}



/*
 **************************************************************************
 * 
 * Function table for UDP PXE functions
 */
struct functab udp_func_tab = {
	UDP_MINNUM, UDP_MAXNUM,
	{{ pxe_udp_open,	sizeof(t_udp_open)},
	 { pxe_udp_close, 	sizeof(t_udp_close)},
	 { pxe_udp_read, 	sizeof(t_udp_read)},
	 { pxe_udp_write, 	sizeof(t_udp_write)}
	}
};

